#include "cgi-search.h"
#include "cgi-srch-fils.h"
#include <ctype.h>
#include <fcntl.h>
#include <locale.h>
#include "srch-version.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <wchar.h>
#include <wctype.h>
/*
*    Defines in _OPTIONAL_ file 'custom.h';
*    CGS_STYLE_SHEET "/Path/File.css" for style sheet.
*    CGS_HTML_FOOT "\nSome_Text\n\n" for additional text in footer.
*    Note: '__has_include' does not work on older C compilers!
*    In which case, you may need the change the stuff below.
*/
#if __has_include ( "custom.h" )
 #include "custom.h"
#endif

/*
*    Form processor for websearch form.
*    Works with cgi-index.
*
*    Copyright (c) 2020 - 2022 R.J. van der Putten, Leiden, Holland, 
*    rob at sput dot nl.
* 
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License version 3 as
*    published by the Free Software Foundation.
* 
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*/

/*    Background colours.    */
char CGS_BG_BROWN[]  = "#EEBB66";
char CGS_BG_RED[]    = "#FFB3B3";
char CGS_BG_ORANGE[] = "#FFDD99";
char CGS_BG_YELLOW[] = "#FFFF99";
char CGS_BG_GREEN[]  = "#AAFFAA";
char CGS_BG_BLUE[]   = "#AAEEFF";
char CGS_BG_VIOLET[] = "#CCC3FF";
char CGS_BG_GREY[]   = "#DDDDDD";
/*  Selection  */
char *bgcolor4[] = { CGS_BG_RED, CGS_BG_YELLOW, CGS_BG_GREEN, CGS_BG_BLUE };
/*  All  */
char *bgcolor8[] = { \
    CGS_BG_BROWN, CGS_BG_RED,  CGS_BG_ORANGE, CGS_BG_YELLOW, \
    CGS_BG_GREEN, CGS_BG_BLUE, CGS_BG_VIOLET, CGS_BG_GREY \
};

/*
    CGS_WRD_SIZ is maximum size of searchword + 1
    CGS_SWRDS_SIZ is maximum nr of searchwords + 1
*/

/*  To make things easy  */
int wlentrs;			/*  Nr of entries in wordlist  */
int wrdfd;			/*  Wordlist FD  */
size_t wlsiz;			/*  Size in bytes of wordlist   */
struct wrdentr *wrdaddr;	/*  Wordlist address after mmap  */
/*  Compact wordlist  */
int wlixfd;			/*  Wordlist index FD  */
struct stridx *wrdidxad;	/*  Wordlist index address  */		
size_t wlidxsiz;		/*  Size in bytes of wordlist index   */
char *cpwladr;			/*  Compact wordlist address after mmap  */

/*  Word to document array  */
unsigned short docs[CGS_SWRDS_SIZ][CGS_DOCS_SIZ];


/*  HTML footer  */
void printfoot(void)
{
    #ifdef CGS_HTML_FOOT
     printf("%s", CGS_HTML_FOOT);
    #endif
    printf("</body>\n");
    printf("</html>\n");
}

/*  HTML error page  */
void printerr(char *s)
{
    printf("<h2>Error</h2>\n\n<p>\n%s<br>\n", s);
    printf("Copyright (c) 2020 - 2022 R.J. van der Putten, Leiden, Holland, ");
    printf("rob at sput dot nl.\n</p>\n\n");

    printfoot();
}

/*  Print the form  */
void printform(struct wrdentr *words)
{
    int i;

    i = 0;

    printf("<form action=\"/cgi-bin/cgi-search\" method=\"POST\">\n");
    printf("<p>\n");
    printf("<input type=\"text\" name=\"search\" size=80 maxlength=%d ", CGS_BUF_MAX);
    printf("value=\"");
    while (words[i].str[0] != 0 && i < CGS_SWRDS_SIZ) {
	printf("%s ", words[i].str);
	i++;
    }
    printf("\"><br>\n<br>\n");
    printf("<input type=\"submit\" value=\"Search\">\n</p>\n</form>\n");
}


/*  Word compare  */
static int compwrd(const void *m1, const void *m2)
{
    struct wrdentr *wrd1, *wrd2;

    /*   Cast from void to struct wrdentr   */
    wrd1 = (struct wrdentr *) m1;
    wrd2 = (struct wrdentr *) m2;
    /*
    *  And then compare;
    *  int strcmp(const char *s1, const char *s2);
    *  Note: strcmp() treats char as unsigned!
    */
    return(strcmp(wrd1->str, wrd2->str));
}

/*  Close wordlist (and index)  */
void closewl(void)
{
    if(cpwladr != NULL) {
	munmap(cpwladr, wlidxsiz);
	cpwladr = NULL;
    }
    if(wrdaddr != NULL) {
	munmap(wrdaddr, wlsiz);
	wrdaddr = NULL;
    }
    if (wlixfd > 0) {
	close(wrdfd);
	wrdfd = -1;
    }
    close(wrdfd);
    wrdfd = -1;
}

/*  Lookup word numbers  */
int lookupwordnrs(struct wrdentr *words)
{
    int foundwrdcnt, i;
    struct stat sb;
    struct wrdentr key, *res;

    foundwrdcnt = 0;
    i = 0;
    wlentrs = 0;
    wrdfd = 0;
    wlsiz = 0;
    memset(&sb,  0, sizeof(sb));
    memset(&key, 0, sizeof(key));
    wrdaddr = NULL;
    res = NULL;

    /*  Open words file  */
    if ((wrdfd = open(CGS_WORDSLIST, O_RDONLY)) < 0) {
	printerr("Can't open words list.");
	exit(1);
    }
    if (fstat(wrdfd, &sb) != 0) {
	printerr("Fstat failed.");
	close(wrdfd);
	wrdfd = -1;
	exit(1);
    }
    wlsiz = (size_t) sb.st_size;
    wlentrs = (int) (wlsiz / sizeof(struct wrdentr));

    /*  Map words file  */
    wrdaddr = mmap(NULL, wlsiz, PROT_READ, MAP_PRIVATE, wrdfd, 0);
    if (wrdaddr == MAP_FAILED) {
	printerr("Mmap failed.");
	close(wrdfd);
	wrdfd = -1;
	exit(1);
    }

    /*  Find word numbers  */
    i = 0;
    while (words[i].str[0] != 0 && i < CGS_SWRDS_MAX) {
	strcpy(key.str, words[i].str);
	/* void *bsearch(const void *key, const void *base, size_t nmemb, \
	  size_t size,int (*compar)(const void *, const void *)); */
	res = bsearch(&key, wrdaddr, wlentrs, sizeof(struct wrdentr), compwrd);
	if (res == NULL) {
	    printf("<p>\nWord %s not found.\n</p>\n\n", words[i].str);
	} else {
	    words[i].num = res->num;
	    foundwrdcnt++;
	}
	i++;
    }

    return(foundwrdcnt);
}

/*  Convert word number to word  */
char * idx2word(int i)
{
    char *word;

    word = NULL;

    /* Note: i = num -1 !!!  */
    if (wrdidxad[i].ptr + wrdidxad[i].len > (int) wlsiz) {
	printerr("Wordlist index mismatch");
	closewl();
	exit(1);
    }
    word = cpwladr + wrdidxad[i].ptr;

    return(word);
}

/*  Lookup a word in compact wordlist  */
int lookupcompword(char *key)
{
    int first, last, middle, ret;
    char *word;

    first = 0;
    ret = 0;
    word = NULL;

    last = wlentrs - 1;
    middle = (first + last) / 2;

    while (first <= last) {
	word = idx2word(middle);
	ret = strcmp(word, key);
	if (ret < 0) {
	    /*  str < key. Examine 2nd half  */
	    first = middle + 1;
	} else if (ret == 0) {
	    /*  str == key; Found, Return number  */
	    return(wrdidxad[middle].num);
	} else {
	    /*  str > key. Examine 1st half  */
	    last = middle - 1;
	}
        middle = (first + last) / 2;
    }

    return(-1);
}

/*  Lookup word numbers in compact list  */
int lookupcompwordnrs(struct wrdentr *words)
{
    int foundwrdcnt, i, res;
    struct stat sb;

    foundwrdcnt = 0;
    i = 0;
    res = 0;
    memset(&sb,  0, sizeof(sb));

    /*  Try to open words index file  */
    if ((wlixfd = open(CGS_WORDS_IDX, O_RDONLY)) < 0) {
	/*  Wordlist index file doesn't exist; Old style  */
	return(lookupwordnrs(words));
    }
    /*  Compact  */
    printf("<!-- Compact wordlist -->\n");
    if (fstat(wlixfd, &sb) != 0) {
	printerr("Fstat failed.");
	close(wlixfd);
	exit(1);
    }
    wlidxsiz = (size_t) sb.st_size;
    wlentrs = (int) (wlidxsiz / sizeof(struct stridx));

    /*  Map words index file  */
    wrdidxad = mmap(NULL, wlidxsiz, PROT_READ, MAP_PRIVATE, wlixfd, 0);
    if (wrdidxad == MAP_FAILED) {
        printerr("Mmap failed.");
        close(wlixfd);
        exit(1);
    }

    /*  Open words file  */
    if ((wrdfd = open(CGS_WORDSLIST, O_RDONLY)) < 0) {
	printerr("Can't open words list.");
	munmap(cpwladr, wlidxsiz);
	close(wlixfd);
	exit(1);
    }
    if (fstat(wrdfd, &sb) != 0) {
	printerr("Fstat failed.");
	close(wrdfd);
	munmap(cpwladr, wlidxsiz);
	close(wlixfd);
	exit(1);
    }
    wlsiz = (size_t) sb.st_size;

    /*  Map words file  */
    cpwladr = mmap(NULL, wlsiz, PROT_READ, MAP_PRIVATE, wrdfd, 0);
    if (cpwladr == MAP_FAILED) {
	printerr("Mmap failed.");
	close(wrdfd);
	munmap(cpwladr, wlidxsiz);
	close(wlixfd);
	exit(1);
    }

    /*  Find word numbers  */
    i = 0;
    while (words[i].str[0] != 0 && i < CGS_SWRDS_MAX) {
	res = lookupcompword(words[i].str);
	if (res < 0) {
	     printf("<p>\nWord %s not found.\n</p>\n\n", words[i].str);
	} else {
	    words[i].num = res;
	    foundwrdcnt++;
	}
	i++;
    }
    
    return(foundwrdcnt);
}

/*  Lookup document numbers  */
int lookupdocs(struct wrdentr *words)
{
    int docfd, entries, i, idxfd, j, k;
    size_t size;
    struct stat sb;
    struct stridx *idxaddr;
    unsigned short *docaddr;
    struct stridx docids[CGS_SWRDS_SIZ];

    docfd = 0;
    entries = 0;
    i = 0;
    idxfd = 0;
    j = 0;
    k = 0;
    docaddr = NULL;
    idxaddr = NULL;
    memset(docids, 0, sizeof(docids));
    memset(&sb, 0, sizeof(sb));

    /*  Open document index index  */
    if ((idxfd = open(CGS_INDEX_IDX,  O_RDONLY)) < 0) {
	printerr("Can't open index.idx");
	closewl();
	exit(1);
    }
    if (fstat(idxfd, &sb) != 0) {
	printerr("Fstat failed.");
	close(idxfd);
	closewl();
	exit(1);
    }
    size = (size_t) sb.st_size;
    entries = (int) (size / sizeof(struct stridx));
    /*  Map document index index  */
    idxaddr = mmap(NULL, size, PROT_READ, MAP_PRIVATE, idxfd, 0);
    if (idxaddr == MAP_FAILED) {
	printerr("Mmap failed.");
	close(idxfd);
	closewl();
	exit(1);
    }

    /*
    *  words is array of struct wrdentr
    *  int num;			Word number; Starts with one!
    *  char str[CGS_STR_SIZ];	Word. Word may be smaller; Use rest as padding
    */

    while (words[i].str[0] != 0 && i < CGS_SWRDS_MAX) {
	if (words[i].num != 0) {
	    /*  Word had been found  */
	    if (words[i].num > entries) {
		printerr("lookupdocs(): Index docids - words mismatch.");
		munmap(idxaddr, size);
		close(idxfd);
		closewl();
		exit(1);
	    }
	    /* void *memcpy(void *dest, const void *src, size_t n); */
	    memcpy(&(docids[j]), &(idxaddr[words[i].num - 1]), sizeof(struct stridx));
	    if (docids[j].num != words[i].num) {
		printerr("lookupdocs(): Index docids - words num mismatch.");
		munmap(idxaddr, size);
		close(idxfd);
		closewl();
		exit(1);
	    }
	    j++;	/*  Nr of words found  */
        }
	i++;
    }
    /*  Done with document index file  */
    munmap(idxaddr, size);
    close(idxfd);

    printf("<!-- Found %d words -->\n", j);

    /*
    *  docids is array of struct stridx
    *
    *  struct stridx {
    *    int num;	;  Number, Starts with one
    *    int ptr;	;  Location in file
    *    int len;	;  Length in bytes!
    *  };
    */

    /*  Open document index  */
    if ((docfd = open(CGS_INDEXLIST, O_RDONLY)) < 0) {
	printerr("Can't open index-list.");
	closewl();
	exit(1);
    }

    if (fstat(docfd, &sb) != 0) {
	printerr("Fstat failed.");
	close(docfd);
	closewl();
	exit(1);
    }
    size = (size_t) sb.st_size;
    /*  Map document index  */
    docaddr = mmap(NULL, size, PROT_READ, MAP_PRIVATE, docfd, 0);
    if (docaddr == MAP_FAILED) {
	printerr("Mmap failed.");
	close(docfd);
	closewl();
	exit(1);
    }

    i = 0;
    k = 0;
    while (i < j && i < CGS_SWRDS_MAX) {
	/*  Read word nr, doc nrs and term null  */
	if (docids[i].ptr + docids[i].len + 2 > size) {
	    printerr("lookupdocs(): Index docids - words mismatch.");
	    munmap(docaddr, size);
	    close(docfd);
	    closewl();
	    exit(1);
	}
	/*
	*    BUG Fix; Check space in docs array!
	*    docs[Word count][Doc count]
	*    Length is 2 x doc count
	*/
	if ((docids[i].len / 2) > CGS_DOCS_MAX) {
	    printerr("lookupdocs(): Too many documents.");
	    close(docfd);
	    closewl();
	    exit(1);
	}
	/* void *memcpy(void *dest, const void *src, size_t n); */
	memcpy(&(docs[k]), &(docaddr[docids[i].ptr / 2]), docids[i].len + 2);
	/*  Document ids in index-list are 16 bit  */
	if (docs[i][0] != (unsigned short) (0xFFFF & docids[i].num)) {
	    printerr("lookupdocs(): Index docs - docids mismatch.");
	    munmap(docaddr, size);
	    close(docfd);
	    closewl();
	    exit(1);
	}
	if (docs[k][1] != 0) {
	    /*  If docs[k][1] is 0; Word was found but there is no document.  */  
	    k++;
	    /*  BUG Fix; Limit counter to CGS_DOCS_MAX  */
	    /*
	    if (k > CGS_DOCS_MAX) {
		printerr("Too many documents.");
		munmap(docaddr, size);
		close(docfd);
		closewl();
		exit(1);
	    }
	    */
	}
	i++;
    }
    /*  Done with document index  */
    munmap(docaddr, size);
    close(docfd);

    return(k);	/*  Nr of words with doc found  */
}

/*  Print abstract from abstracts file  */
int prtabs(int *abstr, struct wrdentr *words, int wrdcnt)
{
    char *word;
    int absptr, bgcolor, i, j, mtchflg;

    absptr = 0;
    bgcolor = 0;
    j = 0;
    mtchflg = 0;
    word = NULL;

    printf("<p>\n");
    i = 1;
    while (i < CGS_ABSTR_SIZ) {
	j = 0;
	mtchflg = 0;
	absptr = abstr[i] - 1;
	/*  Check wlentrs  */
	if(absptr > wlentrs) {
	    printf("<!-- Worlds list - Abstracts list mismatch -->\n</p>\n");
	    return(-1);
	}
	/*  Word  */
	if (cpwladr == NULL) {
	    /*  No compact wordlist  */
	    word = wrdaddr[absptr].str;
	} else {
	    word = idx2word(absptr);
	}
	while (words[j].str[0] != 0 && j < CGS_SWRDS_SIZ) {
	    if (words[j].num != 0 && \
	      strcmp(words[j].str, word) == 0) {
		mtchflg = 1;
		if (wrdcnt > 4)
		    bgcolor = (j & 7);	/*  % 8  */
		else
		    bgcolor = (j & 3);	/*  % 4  */
		break;
	    }
	    j++;
	}
	if (mtchflg == 0) {
	    printf("%s", word);
	} else {
	    if (wrdcnt > 4)
		printf("<span style=\"background-color: %s;\"><b>%s</b></span>", \
		  bgcolor8[bgcolor], word);
	    else
		printf("<span style=\"background-color: %s;\"><b>%s</b></span>", \
		  bgcolor4[bgcolor], word);
	}
	i++;
	if(i > 1 && (i % 16) == 1)
	    printf("\n");
	else
	    printf(" ");
	if(abstr[i] == 0)
	    break;
    }
    printf("\n</p>\n\n");

    return(0);
}

/*  Print results of search  */
int printresults(unsigned short *doc, struct wrdentr *words, int wrdcnt)
{
    int absfd, absflg, entries, i, idxfd, lnkfd;
    size_t absiz, idxsiz, lnksiz;
    struct stat sb;
    struct abstr *absaddr;
    struct stridx *idxaddr;
    char *lnkaddr;

    absflg = 0;
    absfd = 0;
    absiz = 0;
    entries = 0;
    i = 0;
    idxfd = 0;
    idxsiz = 0;
    lnkfd = 0;
    lnksiz = 0;
    absaddr = NULL;	/*  Abstracts  */
    idxaddr = NULL;	/*  Index to links  */
    lnkaddr = NULL;	/*  Links  */
    memset(&sb, 0, sizeof(sb));

    /*  Open links index  */
    if ((idxfd = open(CGS_LNKS_IDX, O_RDONLY)) < 0) {
	printerr("Can't open links.idx");
	closewl();
	exit(1);
    }
    if (fstat(idxfd, &sb) != 0) {
	printerr("Fstat failed.");
	close(idxfd);
	closewl();
	exit(1);
    }
    idxsiz = (size_t) sb.st_size;
    entries = (int) (idxsiz / sizeof(struct stridx));
    /*  Map link index  */
    idxaddr = mmap(NULL, idxsiz, PROT_READ, MAP_PRIVATE, idxfd, 0);
    if (idxaddr == MAP_FAILED) {
	printerr("Mmap failed.");
	close(idxfd);
	closewl();
	exit(1);
    }
    /*  Open links list  */
    if ((lnkfd = open(CGS_LNKSLIST, O_RDONLY)) < 0) {
	printerr("Can't open links-list");
	munmap(idxaddr, idxsiz);
	close(idxfd);
	closewl();
	exit(1);
    }
    if (fstat(lnkfd, &sb) != 0) {
	printerr("Fstat failed.");
	close(lnkfd);
	munmap(idxaddr, idxsiz);
	close(idxfd);
	closewl();
	exit(1);
    }
    lnksiz = (size_t) sb.st_size;
    /*  Map links list  */
    lnkaddr = mmap(NULL, lnksiz, PROT_READ, MAP_PRIVATE, lnkfd, 0);
    if (lnkaddr == MAP_FAILED) {
	printerr("Mmap failed.");
	close(lnkfd);
	munmap(idxaddr, idxsiz);
	close(idxfd);
	closewl();
	exit(1);
    }

    /*  Try abstracts  */
    if (words != NULL) {
	/*  Print abstracts  */
	absflg = 1;
	if ((absfd = open(CGS_ABSTRLIST, O_RDONLY)) < 0) {
	    printf("<!-- %s open failed -->\n", CGS_ABSTRLIST);
	    absflg = 0;
	} else {
	    if (fstat(absfd, &sb) != 0) {
		printf("<!-- %s fstat -->\n", CGS_ABSTRLIST);
		close(absfd);
		absflg = 0;
	    } else {
		absiz = (size_t) sb.st_size;
		absaddr = mmap(NULL, absiz, PROT_READ, MAP_PRIVATE, absfd, 0);
		if (absaddr == MAP_FAILED) {
		    printf("<!-- %s mmap failed -->\n", CGS_ABSTRLIST);
		    close(absfd);
		    absflg = 0;
		}
	    }
	}
    }

    printf("\n<h2>Results</h2>\n\n");

    i = 0;
    while (doc[i] != 0 && i < CGS_DOCS_SIZ) {
	if (doc[i] > entries) {
	    printerr("printresults(): Document - Link mismatch.");
	    munmap(lnkaddr, lnksiz);
	    close(lnkfd);
	    munmap(idxaddr, idxsiz);
	    close(idxfd);
	    closewl();
	    exit(1);
	}
	if (idxaddr[doc[i] - 1].ptr + idxaddr[doc[i] - 1].len > lnksiz) {
	    printerr("printresults(): Document - Link mismatch.");
	    munmap(lnkaddr, lnksiz);
	    close(lnkfd);
	    munmap(idxaddr, idxsiz);
	    close(idxfd);
	    closewl();
	    exit(1);
	}
	printf("<h3>%s</h3>\n\n", lnkaddr + idxaddr[doc[i] - 1].ptr);

	if (absflg != 0) {
	    /* Pint abstract  */
	    prtabs(absaddr[doc[i] - 1].words, words, wrdcnt);
	}
	i++;
    }

    if (absflg != 0) {
	munmap(absaddr, absiz);
	close(absfd);
    }
    munmap(lnkaddr, lnksiz);
    close(lnkfd);
    munmap(idxaddr, idxsiz);
    close(idxfd);
    closewl();

    return(0);
}


int main(void)
{
    int clen, i, j, k, len, m, ret;
    char c, str[CGS_BUF_SIZ], str2[CGS_BUF_SIZ];
    wchar_t wstr[CGS_BUF_SIZ];
    unsigned int u;
    unsigned short doc[CGS_DOCS_SIZ];
    struct wrdentr words[CGS_SWRDS_SIZ];

    c = 0;
    clen = 0;
    i = 0;
    j = 0;
    k = 0;
    len = 0;
    m = 0;
    ret = 0;
    u = 0;
    wlentrs = 0;			/*  Nr of entries in wordlist  */
    wlsiz = 0;				/*  Size of worlist in bytes  */
    wrdfd = 0;				/*  Wordlist FD  */
    wlixfd = 0;				/*  Wordlist index FD  */
    cpwladr = NULL;			/*  Compact worldlist address  */
    wrdaddr = NULL;			/*  Worldlist address  */
    wrdidxad = NULL;			/*  Worldlist index address  */
    memset(doc,    0, sizeof(doc));	/*  Doc number array  */
    memset(docs,   0, sizeof(docs));	/*  2D doc number array  */
    memset(str,    0, CGS_BUF_SIZ);
    memset(str2,   0, CGS_BUF_SIZ);
    memset(words,  0, sizeof(words));	/*  Words and their numbers  */
    memset(wstr,   0, sizeof(wstr));

    setlocale(LC_ALL, "C.UTF-8");

    /*  Print the header  */
    printf("Content-type: text/html\n\n");

    printf("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\"\n");
    printf(" \"http://www.w3.org/TR/html4/strict.dtd\">\n");
    printf("<html>\n<head>\n");
    printf("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n");
    printf("<title>Search</title>\n");
    printf("<link rel=\"shortcut icon\" href=\"/favicon.ico\" type=\"image/x-icon\">\n");
    #ifdef CGS_STYLE_SHEET
     printf("<link rel=\"stylesheet\" type=\"text/css\" href=\"%s\">\n", \
       CGS_STYLE_SHEET);
    #endif
    printf("<!-- Software by R.J. van der Putten, Leiden, Holland, ");
    printf("rob at sput dot nl -->\n");
    printf("<!-- Software version: %s -->\n", cgsversion);
    printf("</head>\n<body>\n\n");

    #ifdef CGS_STYLE_SHEET
     printf("<h1>Search</h1>\n\n");
    #else
     printf("<h1 style=\"text-align: center;\">Search</h1>\n\n");
    #endif

    /*   Check the request method   */
    if (getenv("REQUEST_METHOD") == NULL) {
	printerr("Missing REQUEST_METHOD");
	exit(1);
    }
    if (strcmp("POST", (getenv("REQUEST_METHOD"))) != 0) {
	printerr("Wrong REQUEST_METHOD");
	exit(1);
    }

    /*   Check the Content length   */
    if (getenv("CONTENT_LENGTH") == NULL) {
	printerr("Missing CONTENT_LENGTH");
	exit(1);
    }
    clen = atoi(getenv("CONTENT_LENGTH"));
    if (clen > CGS_BUF_MAX){
	printerr("CONTENT_LENGTH too large");
	exit(1);
    }

    /*   Get the data   */
    fgets(str, CGS_BUF_SIZ, stdin);
    len = strlen(str);
    if (len != clen) {
	printerr("CONTENT_LENGTH mismatch");
	exit(1);
    }

   /*   Check input   */
    if (strncmp (str, "search=", 7) != 0) {
        printerr("Form error\n");
        exit(1);
    }

    /*   Do dehex   */
    i = 7;
    j = 0;
    while (i < clen) {
	c = str[i];
	if (str[i] == '%') {
	    if (isxdigit(str[i + 1]) == 0 || isxdigit(str[i + 2]) == 0) {
	        /*   Not a valid hex value   */
		printerr("Not a valid hex value");
		exit(1);
            }
	    sscanf(str + i + 1, "%2X", &u);
	    c = (char) (0xFF & u);
	    i += 2;
	} else if (str[i] == '+') {
	    c = ' ';
	} else {
	    c = str[i];
	}
	str2[j] = c;
	i++;
	j++;
    }
    str2[j] = 0;

    /*  Force all tolower  */
    /* size_t mbstowcs(wchar_t *dest, const char *src, size_t n); */
    if (mbstowcs(wstr, str2, CGS_BUF_MAX) < 0) {
	printerr("mbstowcs failed");
	exit(1);
    }
    for (i = 0; i < j; i++) {
	if (wstr[i] != 0x2126) {
	    /*  OHM SIGN is not an GREEK CAPITAL LETTER OMEGA  */
	    wstr[i] = towlower(wstr[i]);
	}
    }
    /* size_t wcstombs(char *dest, const wchar_t *src, size_t n); */
    if (wcstombs(str2, wstr, CGS_BUF_MAX) < 0) {
	printerr("wcstombs failed");
	exit(1);
    }

    /*  Fill searchwords array  */
    i = 0;
    j = 0;
    k = 0;
    len = strlen(str2);
    while (i < len && j < CGS_SWRDS_MAX) {
	if (str2[i] < '!' && str2[i] > 0) {
	    str2[i] = 0;
	    if (str2[k] != 0) {
		strncpy(words[j].str, str2 + k, CGS_WRD_MAX);
		words[j].str[CGS_WRD_MAX] = 0;	/*  Terminate  */
		j++;
	    }
	    k = i + 1;
	}
	i++;
    }
    while (str2[k] < '!' && str2[k] > 0 && k < len)
	k++;
    if (j < CGS_SWRDS_MAX && k < len) {
	strncpy(words[j].str, str2 + k, CGS_WRD_MAX);
	words[j].str[CGS_WRD_MAX] = 0;
	j++;
    }

    /*  Find searched word numbers  */
    //lookupwordnrs(words);
    lookupcompwordnrs(words);
    /*  Found word now have their number in words[x].num  */

    /*  Find matching documents  */
    ret = lookupdocs(words);	/*  docs is global  */

    printf("<!-- Found: %d words with docs -->\n", ret);

    /*  Print new form  */
    printform(words);

    /*
    *  Print results
    *  Ret is the number of words found.
    *  docs[x][0] is 16 lower bits of word nr;
    *    This is 0 at 64k wrap!
    *  docs[x][1] is first document number.
    *  If there are no documents for this word docs[x][1] is 0;
    *    This should not happen!
    */
    if (docs[0][1] == 0) {
	printf("<p>\nNo words found.<p>\n");
	printfoot();
	exit(0);
    }

    i = 0;
    j = 0;
    printf("\n <!-- \n");
    for (i = 0; i < ret; i++) {
	printf("%04X ", docs[i][0]);
	for (j = 1; docs[i][j] != 0; j++) {
	    printf("%04X ", docs[i][j]);
	}
	printf("\n");
    }
    printf(" -->\n");

    i = 0;
    if (ret == 1) {
	/*  One word found  */
	printf("<!-- One word found -->\n");
	while (docs[0][i + 1] != 0) {	/* Wrap compat  */
	    doc[i] = docs[0][i + 1];	/*  1st (0) is word nr  */
	    i++;
	    if (i > CGS_DOCS_MAX) {
		printerr("Too many documents.\n");
		exit(1);
	    }
	}
	m = i;
    } else {
	/*  Find common stuff in order to do AND of all words.  */
	/*  docs[word][doc]  */
	i = 1;	/*  docs[x][0] is word nr  */
	m = 0;  /*  Index for docs with ALL words  */
	while (docs[0][i] != 0) {
	    /*  Walk through docs with 1st word  */
	    j = 1;
	    /*  ret is nr of words  */
	    while (j < ret) {
		/*  Walk through other words  */
		k = 1;
		while (docs[j][k] != 0) {
		    /*  Walk through docs with word 'j'  */
		    if (docs[j][k] == docs[0][i]) {
			/*  Doc j,k is doc 'i'  */
			break;	/*  Don't reach docs[j][k] == 0  */
		    }
		    k++;
		    /*
		    *  If at this point docs[j][k] == 0
		    *    there is no match for doc 'i'
		    */
		}
		if(docs[j][k] == 0)
		    break;	/*  Try next 'i'  */
		if (j == ret - 1) {
		    /*  Reached last word; Found  */
		    doc[m] = docs[j][k];
		    m++;
		    if (m > CGS_DOCS_MAX) {
			printerr("Too many documents.\n");
			exit(1);
		    }
		}
		j++;
	    }
	    i++;
	}
    }

    printf("<p>\n%d results\n</p>\n", m);

    /*  Print results  */
    if (words[1].str[0] == 0)
	printresults(doc, NULL, 0);	/*  One word; Don't print abstract  */
    else
	printresults(doc, words, ret);

    /*  Print the footer  */
    printfoot();

    return(0);
}

